import os
import sys
import psutil
import subprocess
import time
import json
import socket

import keyboardsounds.daemon as daemon
from keyboardsounds.profile import Profile
from keyboardsounds.external_api import ExternalAPI


class DaemonManager:
    def __init__(self, lock_file, one_shot = False) -> None:
        """
        Initializes the DaemonManager object with a specified lock file.

        Parameters:
        - lock_file (str): The path to the lock file used to manage the daemon's
                           state.

        Returns:
        - None
        """
        self.__lock_file = lock_file
        self.__proc_info = None
        self.__lock_exists = False
        self.__is_daemon_process = False
        self.__proc = None
        self.__api = None
        self.__one_shot = one_shot
        if not self.__one_shot:
            self.__load_status()

    def __load_status(self):
        if self.__one_shot:
            return
        
        """
        Loads the daemon's current status from the lock file, if it exists, and
        updates internal state accordingly. This includes checking if the lock
        file exists, reading process information, and determining if the current
        process is the daemon process.

        Parameters:
        - None

        Returns:
        - None
        """
        if os.path.isfile(self.__lock_file):
            self.__lock_exists = True
            try:
                with open(self.__lock_file, "r") as f:
                    self.__proc_info = json.load(f)
            except ValueError:
                self.__proc_info = None
                pass
        else:
            self.__proc_info = None

        if self.__proc_info:
            try:
                self.__proc = psutil.Process(self.__proc_info["pid"])
            except psutil.NoSuchProcess:
                self.__proc = None
                pass
        else:
            self.__proc = None

        if self.__proc:
            try:
                proc_name = self.__proc.name()
                current_name = psutil.Process().name()
                self.__is_daemon_process = proc_name == current_name
            except:
                self.__proc = None
                pass
        else:
            self.__is_daemon_process = False

    def status(self, full=False, short=False) -> str:
        """
        Returns the current status of the daemon process. Can provide a simple
        status or a full status with additional details.

        Parameters:
        - full (bool): If True, returns a detailed status string including
                       volume, PID, and profile. If False, returns a simple
                       status string.
        - short (bool): If True, returns a JSON string with the status details.

        Returns:
        - str: The status of the daemon, which can be 'running', 'stale', or
               'free'. If 'full' or 'short' is True, returns a detailed status
                string or JSON string, respectively.
        """
        if full:
            volume = self.get_volume()
            volume_status = f", Volume: {volume}%" if volume is not None else ""
            pid = self.get_pid()
            pid_status = f", PID: {pid}" if pid is not None else ""
            prof = self.get_profile()
            profile_status = f", Profile: {prof}" if prof is not None else ""
            status_text = {"running": "Running", "stale": "Stale"}.get(
                self.status(), "Not running"
            )
            status = f"{status_text}{volume_status}{pid_status}{profile_status}"
            return f"Status: {status}"
        elif short:
            volume = self.get_volume()
            pid = self.get_pid()
            prof = self.get_profile()
            daemon_status = self.status()
            api_port = self.get_api_port()

            user_status = None
            if daemon_status == "running":
                user_status = "Running"
            else:
                user_status = "Not running"

            status = {
                "status": daemon_status,
                "user_status": user_status,
                "volume": volume,
                "pid": pid,
                "profile": prof,
                "api_port": api_port,
                "lock": {
                    "active": self.__lock_exists,
                    "file": os.path.abspath(self.__lock_file),
                },
            }
            return json.dumps(status)
        else:
            self.__load_status()
            if self.__lock_exists:
                if self.__is_daemon_process:
                    return "running"
                else:
                    return "stale"
            else:
                return "free"

    def get_volume(self) -> int:
        """
        Retrieves the current volume level of the daemon if it is running.

        Parameters:
        - None

        Returns:
        - int or None: The volume level of the daemon, or None if the daemon is
                       not running.
        """
        self.__load_status()
        status = self.status()
        if status == "running":
            return self.__proc_info["volume"]
        return None

    def get_pid(self) -> int:
        """
        Retrieves the PID (Process ID) of the daemon if it is running.

        Parameters:
        - None

        Returns:
        - int or None: The PID of the daemon, or None if the daemon is not
                       running.
        """
        self.__load_status()
        status = self.status()
        if status == "running":
            return self.__proc_info["pid"]
        return None

    def get_profile(self) -> str:
        """
        Retrieves the profile name used by the daemon if it is running.

        Parameters:
        - None

        Returns:
        - str or None: The profile name, or None if the daemon is not running.
        """
        self.__load_status()
        status = self.status()
        if status == "running":
            return self.__proc_info["profile"]
        return None

    def get_api_port(self) -> int | None:
        """
        Retrieves the API port used by the daemon if it is running.

        Parameters:
        - None

        Returns:
        - str or None: The API port, or None if the daemon is not running.
        """
        self.__load_status()
        status = self.status()
        if status == "running":
            return int(self.__proc_info["api_port"])
        return None

    def try_stop(self) -> None:
        """
        Attempts to stop the daemon process if it is running or stale. Cleans up
        the lock file if necessary.

        Parameters:
        - None

        Returns:
        - bool: True if the daemon was stopped or cleaned up successfully, False
                if the daemon was already free.
        """
        status = self.status()
        if status == "free":
            return False

        if status == "running":
            self.__proc.kill()
            status = self.status()

        if status == "stale":
            os.unlink(self.__lock_file)

        self.__proc_info = None
        self.__lock_exists = False
        self.__is_daemon_process = False
        self.__proc = None

        return True

    def try_start(self, volume: int, profile: str, debug: bool) -> None:
        """
        Attempts to start the daemon process with the specified volume and
        profile. Stops any existing daemon process before starting a new one.

        Parameters:
        - volume (int): The volume level for the daemon.
        - profile (str): The profile name to be used by the daemon.

        Returns:
        - bool: True if the daemon was started successfully, False if there was
                an error during startup.
        """
        try:
            Profile(profile)
        except ValueError as err:
            print(f"Error: {err}")
            return False

        status = self.status()
        if status == "running":
            self.try_stop()
            status = self.status()

        if status == "stale":
            self.try_stop()

        if debug:
            self.run_daemon(volume, profile, debug=True)
        else:
            subprocess.Popen(
                [sys.argv[0], "start-daemon", str(volume), profile],
                creationflags=subprocess.CREATE_NO_WINDOW if os.name == "nt" else 0,
                start_new_session=True,
            )
            time.sleep(1.0)
        return True

    def update_lock_file(self, volume: int, profile: str):
        lockData = {
            "pid": os.getpid(),
            "volume": volume,
            "profile": profile,
            "api_port": self.__api.port(),
        }
        print(f"updating lock-file with {lockData}")
        with open(self.__lock_file, "w") as f:
            json.dump(lockData, f)

    def capture_daemon_initialization(self):
        """
        Captures and initializes the daemon process based on command-line
        arguments. This method is intended to be called at the start of the
        program.

        Parameters:
        - None

        Returns:
        - bool: True if the daemon was initialized successfully, False if the
                conditions for initialization were not met.
        """
        if len(sys.argv) == 4 and sys.argv[1] == "start-daemon":
            if self.status() == "running":
                return

            volume = 100
            try:
                volume = int(sys.argv[2])
            except:
                pass

            profile = "ios"
            try:
                profile = sys.argv[3]
            except:
                pass

            self.run_daemon(volume, profile, debug=False)
            return True
        return False

    def run_daemon(self, volume: int, profile: str, debug: bool):
        api_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        api_socket.bind(("localhost", 0))
        self.__api = ExternalAPI(api_socket, daemon.on_command)
        self.__api.listen()

        self.update_lock_file(volume, profile)

        if not debug:
            LINE_BUFFERED = 1
            f = open(os.devnull, "w", buffering=LINE_BUFFERED)
            sys.stdout = f
            sys.stderr = f
        daemon.run(self, volume, profile, debug=debug)

    def capture_oneshot(self) -> bool:
        if self.__one_shot and (len(sys.argv) == 3 or len(sys.argv) == 4) and sys.argv[1] == "one-shot":
            pressSound = sys.argv[2]
            releaseSound = None
            if len(sys.argv) == 4:
                releaseSound = sys.argv[3]

            try:
                daemon.one_shot(100, pressSound, releaseSound)
                return True
            except Exception as e:
                print(e)
                return False

        return False


import subprocess
import wave
import os
import io
import random

from typing import Optional

from imageio_ffmpeg import get_ffmpeg_exe
from pynput.keyboard import Key, KeyCode

from keyboardsounds.profile import Profile


class AudioManager:
    def __init__(self, profile: Profile) -> None:
        """
        Initializes the AudioManager with a given profile.

        Parameters:
        - profile (Profile): The profile object containing configuration
                             settings such as audio sources and key mappings.

        The constructor initializes the internal state, loads the sound clips
        based on the provided profile, and sets up the audio manager.
        """
        self.sounds = {}
        self.profile = profile
        self.__one_shot_press_sound = None
        self.__one_shot_release_sound = None
        self.__prime_audio_clips()
        self.__enabled = True

    def set_profile(self, profile: Profile):
        """
        Sets a new profile for the AudioManager.

        Parameters:
        - profile (Profile): The new profile to be set for the AudioManager.

        This method allows for changing the profile associated with the
        AudioManager instance, updating the sound clips and key mappings
        accordingly.
        """
        self.sounds = {}
        self.profile = profile
        self.__prime_audio_clips()

    def __prime_audio_clips(self):
        """
        Primes audio clips based on the profile configuration.

        This private method reads the profile configuration to determine the
        type of audio sources (e.g., video files, individual audio files) and
        prepares the audio clips accordingly. It might involve converting video
        files to audio, extracting specific segments from audio files, and
        organizing them for playback.

        No parameters or return values as it modifies the internal state of the
        AudioManager instance.
        """
        if self.profile.value("profile.type") == "video-extract":
            ffmpeg_exe = get_ffmpeg_exe()
            V_FILE = self.profile.get_child(self.profile.value("profile.video")).get_path()
            A_FILE = f"{V_FILE}.wav"
            subprocess.run(
                [ffmpeg_exe, "-y", "-i", V_FILE, "-f", "wav", "-vn", A_FILE],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            for source in self.profile.value("sources"):
                self.__extract(source["id"], A_FILE, source["start"], source["end"])
            os.unlink(A_FILE)
        elif self.profile.value("profile.type") == "one-shot":
            press = self.profile.value('profile.press')
            release = self.profile.value('profile.release')
            if press is not None:
                self.__extract("one-shot-press", input=press)
                pressBytes = self.sounds["one-shot-press"].getbuffer().tobytes()
                pressAudioData = io.BytesIO(pressBytes)
                self.__one_shot_press_sound = pressAudioData
            if release is not None:
                self.__extract("one-shot-release", input=release)
                releaseBytes = self.sounds["one-shot-release"].getbuffer().tobytes()
                releaseAudioData = io.BytesIO(releaseBytes)
                self.__one_shot_release_sound = releaseAudioData

            
        elif self.profile.value("profile.type") == "files":
            for source in self.profile.value("sources"):
                if isinstance(source["source"], dict):
                    source_id = source["id"]
                    press_loc = source["source"]["press"]
                    press_path = self.profile.get_child(press_loc).get_path()
                    release_loc = source["source"]["release"]
                    release_path = self.profile.get_child(release_loc).get_path()
                    press_id = f"{source_id}__press"
                    release_id = f"{source_id}__release"

                    self.__extract(press_id, press_path)
                    self.__extract(release_id, release_path)

                    press_snd = self.sounds[press_id].getbuffer().tobytes()
                    release_snd = self.sounds[release_id].getbuffer().tobytes()

                    self.sounds[source_id] = {
                        "press": io.BytesIO(press_snd),
                        "release": io.BytesIO(release_snd),
                    }
                    del self.sounds[press_id]
                    del self.sounds[release_id]
                elif type(source["source"]) == str:
                    source_id = source["id"]
                    path = self.profile.get_child(source["source"]).get_path()
                    self.__extract(source_id, path)

    def __extract(self, id, input, start: float = 0.0, end: float = None):
        """
        Extracts and prepares an audio clip from the specified input source.

        Parameters:
        - id (str): An identifier for the audio clip.
        - input (str): The file path of the input audio or video file.
        - start (float, optional): The start time in seconds from which the
                                   audio clip should be extracted.
                                   Defaults to 0.0.
        - end (float, optional):   The end time in seconds till which the audio
                                   clip should be extracted. If None, the clip
                                   is extracted till the end of the file.

        This method handles both audio and video files, extracting the required
        segment and storing it in memory for quick access. The extracted audio
        clip is associated with the provided id.
        """
        if input.endswith(("mp3", "MP3")):
            source = open(input, "rb")
            data = source.read()
            source.close()
            self.sounds[id] = io.BytesIO(data)
        else:
            source = wave.open(input, "rb")
            source.setpos(int(start * source.getframerate()))
            end = end or source.getnframes() / source.getframerate()
            frames = int((end - start) * source.getframerate())
            self.sounds[id] = io.BytesIO()
            dest = wave.open(self.sounds[id], "wb")
            dest.setparams(source.getparams())
            dest.writeframes(source.readframes(frames))
            source.close()
            self.sounds[id].seek(0)

    def get_sound(self, key, action: str = "press") -> Optional[io.BytesIO]:
        """
        Retrieves the sound clip associated with a particular key and action.

        Parameters:
        - key:                    The key for which the sound needs to be
                                  retrieved. Can be an instance of
                                  pynput.keyboard.Key or KeyCode, or a
                                  character.
        - action (str, optional): The type of action, either 'press' or
                                  'release'. Defaults to 'press'.

        Returns:
        - (Optional[io.BytesIO]): A BytesIO object containing the sound clip if
                                  available and AudioManager is enabled;
                                  otherwise, None.

        This method looks up the sound clip based on the provided key and
        action, taking into account any custom mappings defined in the profile.
        If no specific sound is mapped for the key, a default or random sound
        might be returned based on the profile configuration.
        """
        if not self.__enabled:
            return None

        if self.profile.value("keys") is not None:
            if self.profile.value("keys.other") is not None:
                for mapping in self.profile.value("keys.other"):
                    k_val = key.name if isinstance(key, Key) else key.char
                    if k_val in mapping["keys"]:
                        return self.__get_sound(mapping["sound"], action)
            if self.profile.value("keys.default") is not None:
                default_key = self.profile.value("keys.default")
                return self.__get_sound(default_key, action)
        return self.__get_sound(key=None, action=action)

    def get_one_shot_sounds(self) -> list[Optional[io.BytesIO]]:
        return [self.__one_shot_press_sound, self.__one_shot_release_sound]

    def set_enabled(self, enabled: bool) -> None:
        """
        Enables or disables the AudioManager.

        Parameters:
        - enabled (bool): A boolean flag to enable (True) or disable (False)
                          the AudioManager.

        When disabled, the AudioManager will not return any sound clips for key
        events. This method allows for dynamic enabling/disabling of the
        AudioManager at runtime.
        """
        self.__enabled = enabled

    def __get_sound(self, key=None, action: str = "press") -> io.BytesIO:
        """
        A private method to retrieve a sound clip based on a key and action.

        Parameters:
        - key (optional):         The key for which to retrieve the sound. If
                                  None, a random key's sound will be selected.
        - action (str, optional): The type of action, either 'press' or
                                  'release'. Defaults to 'press'.

        Returns:
        - (io.BytesIO): A BytesIO object containing the sound clip.

        This method encapsulates the logic for determining the appropriate sound
        clip based on the key and action, including handling default and random
        sound selection.
        """
        if key is None:
            key = list(self.sounds.keys())
        if type(key) is list:
            return self.__parse_sound(self.sounds[random.choice(key)], action)
        return self.__parse_sound(self.sounds[key], action)

    def __parse_sound(self, sound, action: str = "press") -> io.BytesIO:
        """
        Converts a sound clip into a BytesIO object suitable for playback.

        Parameters:
        - sound:                  The sound clip to be parsed. Can be a direct
                                  audio clip or a dictionary containing 'press'
                                  and 'release' clips.
        - action (str, optional): The type of action, either 'press' or
                                  'release'. Defaults to 'press'.

        Returns:
        - (io.BytesIO): A BytesIO object containing the sound clip ready for
                        playback.

        This method processes the sound clip, ensuring it is in the correct
        format for playback. If the sound clip is a dictionary (containing
        separate clips for key press and release), the appropriate clip is
        selected based on the action.
        """
        if type(sound) is dict:
            return io.BytesIO(sound[action].getbuffer().tobytes())
        elif action == "press":
            return io.BytesIO(sound.getbuffer().tobytes())
        return None



import os
import yaml
import zipfile
import shutil
import tempfile
import requests

from keyboardsounds.root import ROOT
from keyboardsounds.path_resolver import PathResolver
from keyboardsounds.profile_validation import validate_profile
from keyboardsounds.profile_builder import CliProfileBuilder

PROFILES_REMOTE_URL = "https://api.github.com/repos/nathan-fiscaletti/keyboardsounds/contents/keyboardsounds/profiles?ref=master"
PROFILE_REMOTE_URL = "https://api.github.com/repos/nathan-fiscaletti/keyboardsounds/contents/keyboardsounds/profiles/{name}?ref=master"
PROFILE_DETAIL_URL = "https://raw.githubusercontent.com/nathan-fiscaletti/keyboardsounds/master/keyboardsounds/profiles/{name}/profile.yaml"

def OneShotProfile(press_sound: str = None, release_sound: str = None):
    return Profile("one-shot", one_shot=True, one_shot_press_sound=press_sound, one_shot_release_sound=release_sound)
class Profile(PathResolver):
    def __init__(self, name: str, one_shot: bool = False, one_shot_press_sound:str = None, one_shot_release_sound:str = None):
        super().__init__(os.path.join(ROOT, "profiles", name))
        
        self.name = name
        self.__one_shot = one_shot
        self.__one_shot_press_sound = one_shot_press_sound
        self.__one_shot_release_sound = one_shot_release_sound
        self.__validate()

    def __validate(self):
        
        # Check for Oneshot
        
        if self.__one_shot:
            self.__data = {
                'profile': {
                    'type': "one-shot",
                    'press': self.__one_shot_press_sound,
                    'release': self.__one_shot_release_sound
                }
            }
            return

        # Validate File Structure

        if not os.path.isdir(self.root):
            raise ValueError(f"Profile '{self.name}' does not exist")
        if not os.path.isfile(self.get_child("profile.yaml").get_path()):
            raise ValueError(
                f"Profile '{self.name}' is corrupted. Missing profile.yaml."
            )

        with open(self.get_child("profile.yaml").get_path(), "r") as f:
            data = yaml.safe_load(f)
            self.__data = validate_profile(self, data)

    def value(self, key: str):
        value = self.__data
        map = key.split(".")
        for key in map:
            if key not in value:
                return None
            value = value[key]
        return value

    def data(self):
        return self.__data

    def metadata(self):
        name = self.__data["profile"]["name"]
        author = (
            self.__data["profile"]["author"]
            if "author" in self.__data["profile"]
            else None
        )
        description = (
            self.__data["profile"]["description"]
            if "description" in self.__data["profile"]
            else None
        )
        return {"name": name, "author": author, "description": description}

    def export(self, output: str):
        # export profile to zip file
        CliProfileBuilder(self.root, output).save()

    @classmethod
    def list(cls):
        names = [
            f.name for f in os.scandir(os.path.join(ROOT, "profiles")) if f.is_dir()
        ]
        return [Profile(name, one_shot=False, one_shot_press_sound=None, one_shot_release_sound=None) for name in names]

    @classmethod
    def remove_profile(cls, name: str):
        output_path = os.path.join(ROOT, "profiles", name)
        if os.path.isdir(output_path):
            shutil.rmtree(output_path)
            print("Profile removed.")
        else:
            print("Profile does not exist.")

    @classmethod
    def list_remote_profiles(cls):
        response = requests.get(PROFILES_REMOTE_URL)
        if response.status_code != 200:
            raise ValueError("Failed to fetch remote profiles.")
        profile_names = [f["name"] for f in response.json() if f["type"] == "dir"]

        result = []
        # Retrieve `profile.yaml` for the profile
        for profile_name in profile_names:
            response = requests.get(PROFILE_DETAIL_URL.format(name=profile_name))
            if response.status_code != 200:
                raise ValueError(f"Failed to fetch remote profile '{profile_name}'.")
            # parse the yaml file
            profile_data = yaml.safe_load(response.text)
            result.append(profile_data["profile"])
        return result

    @classmethod
    def download_profile(cls, name: str):
        print(f"Retrieving meta-data for profile '{name}'...")
        response = requests.get(PROFILE_REMOTE_URL.format(name=name))
        if response.status_code != 200:
            raise ValueError(f"Failed to fetch remote profile '{name}'.")

        print(f"Downloading profile '{name}'...")
        # Get the profile data
        profile_data = response.json()

        # Create a temporary working directory
        tmpdir = tempfile.mkdtemp()

        # Download the profiles files into the directory
        for file in profile_data:
            if file["type"] == "file":
                response = requests.get(file["download_url"])
                with open(os.path.join(tmpdir, file["name"]), "wb") as f:
                    f.write(response.content)

        # Compress the contents of the directory into a zip file (stored in the temporary directory)
        output_path = os.path.join(tmpdir, f"{name}.zip")
        with zipfile.ZipFile(output_path, "w") as zip_ref:
            for file in os.listdir(tmpdir):
                file_path = os.path.join(tmpdir, file)
                if file_path != output_path:
                    zip_ref.write(file_path, file)

        # Add the profile to the local profiles directory
        err = None
        try:
            Profile.add_profile(output_path)
        except ValueError as e:
            err = e
        finally:
            # Clean up the temporary directory
            shutil.rmtree(tmpdir)
            if err:
                raise err

    @classmethod
    def add_profile(cls, path: str):
        input_path = os.path.abspath(path)
        tmpdir = tempfile.mkdtemp()

        with zipfile.ZipFile(input_path, "r") as zip_ref:
            zip_ref.extractall(tmpdir)

        profile_data = None
        if os.path.isfile(os.path.join(tmpdir, "profile.yaml")):
            with open(os.path.join(tmpdir, "profile.yaml"), "r") as f:
                profile_data = yaml.safe_load(f)

        if (
            not profile_data
            or "profile" not in profile_data
            or "name" not in profile_data["profile"]
        ):
            shutil.rmtree(tmpdir)
            raise ValueError("Profile is corrupted. Missing or invalid profile.yaml.")
        name = profile_data["profile"]["name"]

        print(f"Importing profile from '{input_path}'...")
        output_path = os.path.join(ROOT, "profiles", name)

        shutil.move(tmpdir, output_path)
        try:
            Profile(name)
            print(f"Successfully imported profile '{name}'.")
        except ValueError as e:
            print(f"Profile '{name}' is invalid. Deleting...")
            shutil.rmtree(output_path)
            print(f"Error: {e}")
            return

    @classmethod
    def create_profile(cls, path: str, name: str):
        if name is None:
            print("Please specify a name for the profile to create.")
            return

        # If no directory is provided, use the current directory and the name
        if path is None:
            dirNameSafeName = name.replace(" ", "_").lower()
            path = os.path.join(os.getcwd(), dirNameSafeName)

        # Make sure the directory doesn't already exist
        if os.path.isdir(path):
            print(f"Directory '{path}' already exists.")
            return

        # Create the directory
        try:
            os.makedirs(path)
        except Exception as e:
            raise ValueError(f"Failed to create directory '{path}'. Error: {e}")

        # Create the profile.yaml file using the template
        profile_file = os.path.join(path, "profile.yaml")
        with open(profile_file, "w") as f:
            with open(
                os.path.join(ROOT, "profiles", "profile.template.yaml"), "r"
            ) as template_file:
                template = template_file.read()
                template = template.replace("{{ name }}", name)
                f.write(template)

        print("")
        print(f"Profile '{name}' created successfully.")
        print("")
        print(f"    Profile Directory: {path}")
        print("")
        print("      - Edit the profile.yaml file to customize the profile.")
        print("      - Add audio files to the directory.")
        print("")
        print(
            "    Use the interactive builder to add sources, keys, and default key mappings."
        )
        print("")
        print(f"        kbs build-profile -path {path}")
        print("")
        print("    Build your profile using the following command:")
        print("")
        print(f"        kbs build-profile -path {path} -o {name}.zip")
        print("")

    @classmethod
    def export_profile(cls, name: str, output: str):
        if name is None:
            print("Please specify a name for the profile to export.")
            return

        try:
            profile = Profile(name)
        except Exception as e:
            print(f"Failed to find profile '{name}'. Error: {e}")
            return

        profile.export(output)



import os

class PathResolver:
    def __init__(self, root: str):
        self.root = root

    def get_child(self, child: str):
        return PathResolver(os.path.join(self.root, child))
    
    def get_path(self):
        return self.root